/**********************************************************************
 * main.cc
 * Copyright (c) 2003 Virtual Werder
 *
 * implements:
 *   function main()
 *   function showHelp()
 *
 * requires:
 *
 * Created on: 05.05.2003 by pagra
 * Maintainer: pagra
 * Group:
 * Co-authors:
 *
 * Changelog:
 *
 **********************************************************************/

/** \mainpage VirtualWerderCoach
 *
 * \section intro Introduction
 *
 * This is the documentation for the implementation of the online coach for the Robocup Soccer Server.\n
 * Created by VirtualWerder (University Of Bremen, Robocup II Project).
 *
 */
 
/** @defgroup Main Main Classes & Functions
 *  This group contains all main-coach related classes & functions
 *  @{
 */

#include <iostream>
#include <stdlib.h>
#include <pthread.h>
#include <errno.h>
#include <unistd.h>

#include "debug.hh"
#include "ConfigFile.hh"
#include "Coach.hh"
#include "WorldModel.hh"


// forward declarations
void showHelp();

void *think (void *p);


/**
 This is the main method of the Coach-executable
 First it creates a thread, which is analysing the worldmodel
 Then starts the coach, which receives messages from the server, parses them, 
 distributes heterogenous player-types, analyses the distribution in the other team.
*/
int main(int argc, char** argv){
	
  if( argc >= 5 ) {
    showHelp();
    exit(-1);
  }
	
  if( !ConfigFile::parseConfigFile( "coach.conf" ) ) {
    cerr << "Unable to parse <coach.conf>!" << endl;
    exit(-1);
  }

  char *serverHost = NULL;
  char *teamname   = NULL;
  int   serverPort = 0;

  // if commandline arguments given (always same order), use them
  switch(argc){
  case 4:
    serverPort = atoi( argv[3] );
  case 3:
    teamname   = strdup(argv[2]);
  case 2:
    serverHost = strdup(argv[1]);
  default:
    break;
  }

  // if not given as argument, take the value of the ConfigFile
  if ( serverHost == NULL ) serverHost = strdup( ConfigFile::serverHost.c_str() );
  if ( teamname   == NULL ) teamname   = strdup( ConfigFile::teamname.c_str() );
  if ( serverPort <= 0 )    serverPort = ConfigFile::serverPort;
  	
	
	
 	// This is the coach's WorldModel. It is shared between the coach and the think-thread.
	WorldModel *worldmodel = new WorldModel();
	
	pthread_t thinker;
	pthread_create(&thinker, NULL, think, (void*)worldmodel);
	
  Coach coach = Coach( worldmodel );
  coach.init( serverHost, serverPort, teamname );                                                                         
  coach.run();
	
}



/**
  Think-thread.
	It starts calculations in the worldmodel.
	@param *p void pointer to the worldmodel
 */
void *think (void *p) {
  WorldModel *worldmodel = (WorldModel*) p;
  int lastcycle = 0;
  int avgSkipCycles = 0; // cycles to skip for the calculation of the average positions of the players and the ball
  
  worldmodel->setFormation(ConfigFile::standardFormation);
  
  while (true) {
    
    if (worldmodel->time != lastcycle) {
      
      if (ConfigFile::usePlayerTypeDetection) {
        if (ConfigFile::usePTD_InertiaMoment)
          worldmodel->calcInertiaMoment();
        if (ConfigFile::usePTD_KickableMargin)
          worldmodel->calcKickableMargin();
        if (ConfigFile::usePTD_PlayerSpeedMax)
          worldmodel->calcPlayerSpeedMax();
        if (ConfigFile::usePTD_Acceleration)
          worldmodel->calcAcceleration();
        if (ConfigFile::usePTD_Decay)
          worldmodel->calcDecay();
      }
      
      if (ConfigFile::useOffsideXDetection)
        worldmodel->calcOtherDefenseLineX();
      
      if (ConfigFile::useFormationDetection) {
        if (avgSkipCycles <= 0) 
          avgSkipCycles = worldmodel->updateFormationDetection();
        else 
          avgSkipCycles--;
        if (worldmodel->time % 25 == 0) worldmodel->determineFormation();
      }
      
    }
    
    if (ConfigFile::usePlayerTypeDetection){
      if (worldmodel->time % 50 == 0 && 
                worldmodel->time != lastcycle && 
                worldmodel->time != 0) {
        printf ("\n\n,--------------------------------------------.\n");
        printf (    "| The Coach is pondering for cycle %04i ...  |\n", worldmodel->time);
        printf (    "`--------------------------------------------'--------------.\n");
        worldmodel->determineHeterogenousPlayers();
      }
    }
    
    lastcycle = worldmodel->time;
    usleep(10);
  }
  pthread_exit(0);
}


void showHelp() {
  cout << "usage: coach <host <teamname <port>>>" << endl;
}

/** @} */ // end of Main
