/******************************************************************************
 * WorldModel.hh
 * Copyright (c) 2003 Virtual Werder
 *
 *
 * implements:
 *   class WorldModel
 *
 * requires:
 *
 *
 * Created on: 21.01.2004 by aw
 * Maintainer: aw, mangels
 * Group:
 * Co-authors:
 *
 * $Id: WorldModel.hh,v 1.3 2004/06/24 09:30:39 aw Exp $
 *****************************************************************************/


#ifndef WORLDMODEL_HH
#define WORLDMODEL_HH

#include <vector>
#include "common.hh"
#include "WorldModelState.hh"
#include "ParsedMsg.hh"
#include "ParsedObjects.hh"
#include "PlayerType.hh"
//#include "Player.hh"


using namespace std;


class WorldModelState;

/** @defgroup WorldModel Worldmodel Classes
 *  This group contains all worldmodel related classes
 *  @{
 */


/**
 * This class represents the worldmodel of the coach.
 * @brief class WorldModel in group WorldModel
 */
class WorldModel : public WorldModelState {
private:
  /**
     Vector of WorldModelStates (History of past WorldModels).
     @todo Replace it with another storage-method, e.g. a ringbuffer,
     to have a better memory-performance
  */
  vector<WorldModelState> wmstates;
  
  /**
   * Pointer to Vector of PlayerTimes, comming from the server.
   */
  vector<PlayerType> *playerTypes;
	
  /**
   * Calculated MaxSpeed of the playertypes
   */
  float ptPlayerSpeedMax;
	
  /**
   * Calculated max kickable margin of the playertypes
   */
  float ptKickableMarginMax;
  
  int balanceCount;

  /**
   * Min Value of the x position of the other team
   * for offside calculation
   */
  float otherDefenseLineX;
  
  /**
   * determined formation of the other team
   * internal formation id!!!
   * defined in common.hh
   */
  CoFormation otherFormation;
  
  /**
   * Formation of ourTeam
   * formation id from formation.conf
   */
  unsigned int ourFormation;
  
  /**
   * determined tactic to use
   * defined in common.hh
   */
  CoTactic ourTactic;
  
  /**
   * goal difference
   */
  int score;
  
public:
  /**
   * This is the constructor. It takes no parameters.
   */
  WorldModel();
	
  /**
   * The destructor.
   * @todo to be implemented
   */
  ~WorldModel();

  /**
     This method determines the heterogenous players by the available values from the following calculations
  */
  void determineHeterogenousPlayers();
	
  /**
     This method ... playmode ... into worldmodelstate ...
  */
  void setPlaymode(unsigned int time, Playmode pm){
    vector<WorldModelState>::iterator i;
    for (i = wmstates.begin(); i != wmstates.end(); i++){
      if (i->time == (int)time){
	i->playmode = pm;
	printf("Setting playmode %i to cycle %i\n", pm, time);
	return;
      }
    }
    printf("Couldn't find cycle %i!\n", (int)time);
  }

  /**
   * sets the playmode of the current cycle
   * changes the score of the game
   */
  void setPlaymode (Playmode pm) {
    playmode = pm;
    if (playmode == PM_TheirGoal) 
      score--;
    if (playmode == PM_OurGoal)
      score++;
  }
  
  /**
   * Sets our formation to the given id.. initially set to default in coach.conf
   * @param unsigned int formation
   */
  void setFormation(unsigned int fm) {
    ourFormation = fm;
#if defined (DEBUG_DETERMINE_FORMATION)
    printf ("Setting our formation to id %i\n", ourFormation);
#endif  
  }
  
  int getFormation() const { return ourFormation; }
    
  /**
     This method calculates the maximum player speed of each player over the past WorldModelStates.
  */
  void calcPlayerSpeedMax();
	
  /**
     This method calculates the kickable margin for each player over the past WorldModelStates.
     To find out, if the ball direction has changed, we are using this simple algorithm:
     <pre>
     | ----------- l1 ----------|
     | --- l2 ------- | -- l3 --|
     .________________._________.
     p1               p2       p3

     l1 = l2 + l3 // the distance wouldn't have been changed
     p2              
     .----------------.
     p1                \
     \
     ` p3
     l1 < l2 + l3 // the distance have been changed.
     </pre>
     additionally the value of l2 + l3 is being multiplied with 0.99, to cut out calculation errors, or bad ball behaviour
  */
  void calcKickableMargin();
	
	
  /**
     This method calculate the inertia moment of every player
  */
  void calcInertiaMoment();
  /**
     This method calculate the decay of every player
  */
  void calcDecay(); 
  /**
    This method calculates the acceleration of each player
   */
  void calcAcceleration();

  void calcOtherDefenseLineX();

  /**
     this method integrates a SeeGlobalMsg into the current WorldModel.
     @param msg A See Global Message, constructed by the Parser.
  */
  void integrateSeeGlobalMsg (SeeGlobalMsg *msg);
   
  /**
     this method integrates a ChangePlayerTypeMsg into the current WorldModel.
     @param msg a ChangePlayerTypeMsg, constructed by the Parser.
  */
  void integrateChangePlayerTypeMsg (ChangePlayerTypeMsg *msg);
  
  /**
     Saves the recent WorldModelState
     @todo replace the vector, by a better container, like a ringbuffer.
  */
  void saveState(Playmode pm);
	
  /**
     This method sets the playertypes to the parsed playertypes by the coach.
     @param pt pointer to the playertypes.
  */
  void setPlayerTypes (vector<PlayerType> *pt);

  /**
   * This method calculates the max values of each (not really each) player type ability
   */
  void getPTMaxValues ();


  float getOtherDefenseLine () { return otherDefenseLineX; };
  // void checkDuplicatePlayerTypes();
  
  /**
   * Ensure that there are max three players per type
   * calls itself recursive
   */
  bool balancePlayerTypeDetermination();
  
  /**
   * simply prints the predicted player types 
   */ 
  void printPlayerTypePrediction();
  
  /**
   * Updates the average x/y values for each player
   * if we are in a valid playmode 
   * @return int cycles to skip, due to a 'bad' playmode
   */
  int updateFormationDetection();
  
  void determineFormation();
};

/** @} */ // end of WorldModel

#endif //WORLDMODEL_HH

